# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ==============================
# utils.py
# Description: Utility functions
# ==============================

import os
import json


def load_config_file(path: str) -> dict:
    """Load a JSON configuration file from the specified path and return it as a dictionary."""
    try:
        with open(path, 'r') as f:
            config_dict = json.load(f)
        return config_dict

    except FileNotFoundError:
        print(f"Error: The file '{path}' does not exist.")
        return None
    except json.JSONDecodeError as e:
        print(f"Error decoding JSON in '{path}': {e}")
        # Handle other potential JSON decoding errors here
        return None
    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        # Handle other unexpected errors here
        return None


def load_json_as_list(input_file: str) -> list:
    """Load a JSON file as a list of dictionaries."""
    res = []
    with open(input_file, 'r') as f:
        lines = f.readlines()
    for line in lines:
        d = json.loads(line)
        res.append(d)
    return res


def create_directory_for_file(file_path) -> None:
    """Create the directory for the specified file path if it does not already exist."""
    directory = os.path.dirname(file_path)
    if not os.path.exists(directory):
        os.makedirs(directory)



from .gen_model import huggingface_model
from .assembly_models import assembly_gpt_oss, assembly_qwen3, assembly_gemini

__all__ = ["huggingface_model", "assembly_gpt_oss", "assembly_qwen3", "assembly_gemini"]


from .gen_model import huggingface_model
from .assembly_models import assembly_qwen3
def main():
    gen_model:huggingface_model = assembly_qwen3()
    prompt = "Explain the theory of relativity in simple terms."
    messages = [
        {"role": "user", "content": prompt}
    ]

    watermarked_text = gen_model.generate(messages, do_watermark=True, max_new_tokens=200)
    print(watermarked_text)
    print(gen_model.detect_watermark(watermarked_text))
    print("----------")
    non_watermarked_text = gen_model.generate(messages, do_watermark=False, max_new_tokens=200)
    print(non_watermarked_text)
    print(gen_model.detect_watermark(non_watermarked_text))
    print("----------")
    
    


if __name__ == "__main__":
    main()



from .gen_model import gpt_oss, qwen3
from .gen_model import gemini

from .watermark.auto_watermark import AutoWatermark
from .watermark.transformers_config import TransformersConfig


def assembly_gpt_oss(size="20b", watermark_scheme = "KGW"):
    assert size in ["20b", "120b"], "The size for the gpt oss model should be chosen from 20b or 120b"
    gen_model = gpt_oss(f"openai/gpt-oss-{size}")
    
    transformers_config = TransformersConfig(
        model=gen_model.model,
        tokenizer=gen_model.tokenizer,
        vocab_size=201088,
        device="cuda",
        max_new_tokens=1024,
        min_length=32,
        do_sample=True,
        no_repeat_ngram_size=4
    )
    Watermark = AutoWatermark.load(watermark_scheme, algorithm_config=f'watermark/{watermark_scheme}.json',transformers_config=transformers_config)
    
    gen_model.load_watermark(Watermark)
    
    return gen_model
    
def assembly_qwen3(size="8B", watermark_scheme = "KGW", enable_thinking=False):
    assert size in ["4B",  "8B", "14B", "32B"], "The size for the qwen3 model should be chosen from 8B or 14B"
    gen_model = qwen3(f"Qwen/Qwen3-{size}", enable_thinking=enable_thinking)
    
    transformers_config = TransformersConfig(
        model=gen_model.model,
        tokenizer=gen_model.tokenizer,
        vocab_size=151936,
        device="cuda",
        max_new_tokens=1024,
        min_length=32,
        do_sample=True,
        no_repeat_ngram_size=4
    )
    Watermark = AutoWatermark.load(watermark_scheme, algorithm_config=f'watermark/{watermark_scheme}.json',transformers_config=transformers_config)
    
    gen_model.load_watermark(Watermark)
    
    return gen_model

def assembly_gemini(model_name="gemini-2.5-flash", watermark_scheme = "KGW"):
    pesudo_model = qwen3(f"Qwen/Qwen3-4B", enable_thinking=False)
    
    transformers_config = TransformersConfig(
        model=pesudo_model.model,
        tokenizer=pesudo_model.tokenizer,
        vocab_size=151936,
        device="cuda",
        max_new_tokens=1024,
        min_length=32,
        do_sample=True,
        no_repeat_ngram_size=4
    )
    Watermark = AutoWatermark.load(watermark_scheme, algorithm_config=f'watermark/{watermark_scheme}.json',transformers_config=transformers_config)
    
    gen_model = gemini(model_name=model_name)
    Watermark = AutoWatermark.load(watermark_scheme, algorithm_config=f'watermark/{watermark_scheme}.json',transformers_config=transformers_config)
    
    gen_model.load_watermark(Watermark)
    return gen_model


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ===============================================================
# base.py
# Description: This is a generic watermark class that will be 
#              inherited by the watermark classes of the library.
# ===============================================================

from typing import Union
from .utils import load_config_file
from .transformers_config import TransformersConfig
# from visualize.data_for_visualization import DataForVisualization


class BaseConfig:
    """Base configuration class for watermark algorithms."""

    def __init__(self, algorithm_config_path: str, transformers_config: TransformersConfig, *args, **kwargs) -> None:
        """
        Initialize the base configuration.

        Parameters:
            algorithm_config (str): Path to the algorithm configuration file.
            transformers_config (TransformersConfig): Configuration for the transformers model.
            **kwargs: Additional parameters to override config values
        """
        # Load config file
        self.config_dict = load_config_file(f'config/{self.algorithm_name()}.json') if algorithm_config_path is None else load_config_file(algorithm_config_path)
            
        # Update config with kwargs
        if kwargs:
            self.config_dict.update(kwargs)
            
        # Load model-related configurations
        self.generation_model = transformers_config.model
        self.generation_tokenizer = transformers_config.tokenizer
        self.vocab_size = transformers_config.vocab_size
        self.device = transformers_config.device
        self.gen_kwargs = transformers_config.gen_kwargs
        self.transformers_config = transformers_config
        
        # Initialize algorithm-specific parameters
        self.initialize_parameters()
        
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters. Should be overridden by subclasses."""
        raise NotImplementedError
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name. Should be overridden by subclasses."""
        raise NotImplementedError



class BaseWatermark:
    def __init__(self, algorithm_config: str | BaseConfig, transformers_config: TransformersConfig, *args, **kwargs) -> None:
        pass

    def generate_watermarked_text(self, prompt: str, *args, **kwargs) -> str: 
        pass

    def generate_unwatermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate unwatermarked text."""
        
        # Encode prompt
        encoded_prompt = self.config.generation_tokenizer(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)
        # Generate unwatermarked text
        encoded_unwatermarked_text = self.config.generation_model.generate(**encoded_prompt, **self.config.gen_kwargs)
        # Decode
        unwatermarked_text = self.config.generation_tokenizer.batch_decode(encoded_unwatermarked_text, skip_special_tokens=True)[0]
        return unwatermarked_text

    def detect_watermark(self, text:str, return_dict: bool=True, *args, **kwargs) -> Union[tuple, dict]:
        pass

    def get_data_for_visualize(self, text, *args, **kwargs) -> None: # Not Needed
        pass






# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ===============================================================
# auto_config.py
# Description: This is a generic configuration class that will be 
#              inherited by the configuration classes of the library.
# ===============================================================


import importlib
from typing import Dict, Optional, Any
from .transformers_config import TransformersConfig

CONFIG_MAPPING_NAMES = {
    'KGW': 'watermark.kgw.KGWConfig',
    'Unigram': 'watermark.unigram.UnigramConfig',
    'SWEET': 'watermark.sweet.SWEETConfig',
    'UPV': 'watermark.upv.UPVConfig',
    'SIR': 'watermark.sir.SIRConfig',
    'XSIR': 'watermark.xsir.XSIRConfig',
    'Unbiased': 'watermark.unbiased.UnbiasedConfig',
    'DIP': 'watermark.dip.DIPConfig',
    'EWD': 'watermark.ewd.EWDConfig',
    'EXP': 'watermark.exp.EXPConfig',
    'EXPGumbel': 'watermark.exp_gumbel.EXPGumbelConfig',
    'EXPEdit': 'watermark.exp_edit.EXPEditConfig',
    'ITSEdit': 'watermark.its_edit.ITSEditConfig',
    'SynthID': 'watermark.synthid.SynthIDConfig',
    'TS': 'watermark.ts.TSConfig',
    'PF': 'watermark.pf.PFConfig',
    'MorphMark': 'watermark.morphmark.MorphMarkConfig',
}

def config_name_from_alg_name(name: str) -> Optional[str]:
    """Get the config class name from the algorithm name."""
    if name in CONFIG_MAPPING_NAMES:
        return CONFIG_MAPPING_NAMES[name]
    else:
        raise ValueError(f"Invalid algorithm name: {name}")

class AutoConfig:
    """
    A generic configuration class that will be instantiated as one of the configuration classes
    of the library when created with the [`AutoConfig.load`] class method.

    This class cannot be instantiated directly using `__init__()` (throws an error).
    """

    def __init__(self):
        raise EnvironmentError(
            "AutoConfig is designed to be instantiated "
            "using the `AutoConfig.load(algorithm_name, **kwargs)` method."
        )

    @classmethod
    def load(cls, algorithm_name: str, transformers_config: TransformersConfig, algorithm_config_path=None, **kwargs) -> Any:
        """
        Load the configuration class for the specified watermark algorithm.

        Args:
            algorithm_name (str): The name of the watermark algorithm
            transformers_config (TransformersConfig): Configuration for the transformers model
            algorithm_config_path (str): Path to the algorithm configuration file
            **kwargs: Additional keyword arguments to pass to the configuration class

        Returns:
            The instantiated configuration class for the specified algorithm
        """
        config_name = config_name_from_alg_name(algorithm_name)
        if config_name is None:
            raise ValueError(f"Unknown algorithm name: {algorithm_name}")
            
        module_name, class_name = config_name.rsplit('.', 1)
        module = importlib.import_module(module_name)
        config_class = getattr(module, class_name)
        if algorithm_config_path is None:
            algorithm_config_path = f'config/{algorithm_name}.json'
        config_instance = config_class(algorithm_config_path, transformers_config, **kwargs)
        return config_instance